'use strict';

const VERSION = '0.21';

const registrationDataLocalStorageKey = 'ownid-registration-data';
const enrollCredentialLocalStorageKey = 'ownid-enroll-credential';
const pageVisitLocalStorageKey = 'ownid-page-visited';
const authNotificationShownStorageKey = 'ownid-auth-notification-shown';
const loggedInStorageKey = 'ownid-logged-in';
const ownidSdkStorageKey = 'ownid-s';

const mkorsSiteAlias = 'mkors';
const defaultSiteAlias = 'default';
const ownIdWidgetElementType = 'ownid-fingerprint-button-widget';
const borderColorCssProperty = '--ownid-button-widget-border-color';
let originalBorderColor = null;

/**
 * log
 * @param {string} message - .
 * @param {any[]} params - .
 */
function log(message, ...params) {
    console.log('ownid', message, ...params);
}

/**
 * logError
 * @param {string} message - .
 * @param {any[]} params - .
 */
function logError(message, ...params) {
    console.error('ownid', message, ...params);
}

/**
 * isMKors
 * @returns {boolean} - .
 */
function isMKors() {
    return window.mkorsData !== undefined || window.location.hostname.includes('michaelkors.');
}

/**
 * postAnalyticsEvent
 * @param {string} eventName - .
 * @param {Object} [additionalProperties={}] - .
 */
function postAnalyticsEvent(eventName, additionalProperties = {}) {
    if (isMKors()) {
        const props = additionalProperties || {};
        if (window.ownIdLocale) {
            props.locale = window.ownIdLocale;
        }
        $('body')
            .trigger(`event:${eventName}`, props);
    }
}

/**
 * clearElementsErrors
 * @param {HTMLElement[]} [elements=[]] - An array of HTML elements.
 */
function clearElementsErrors(elements = []) {
    const DELAY_MS = 500;
    setTimeout(() => {
        try {
            if (!elements) {
                return;
            }
            elements.forEach(element => {
                if (!element) {
                    return;
                }
                element.classList.remove('is-invalid');
                let errorElement = document.getElementById(`${element.id}-error`);
                if (errorElement) {
                    errorElement.remove();
                } else {
                    errorElement = document.getElementById(`${element.id}-error`.replace(/(\*)-form-(\*)/, 'form-$2'));
                    if (errorElement) {
                        errorElement.remove();
                    }
                }
            });
            Array.from(
                document.querySelectorAll('div[role="alert"]')
            )
                .filter(element => element.parentElement.tagName.toLowerCase() === 'form')
                .forEach(element => element.remove());
        } catch (e) {
            logError('Failed clearing elements errors', e);
        }
    }, DELAY_MS);
}

/**
 * setLastUser
 * @param {string} loginId - .
 */
function setLastUser(loginId) {
    log(`Setting last user ${loginId}`);
    const dataStr = localStorage.getItem(ownidSdkStorageKey);
    const data = dataStr ? JSON.parse(dataStr) : { lastUser: {} };
    if (!loginId) {
        delete data.lastUser.loginId;
    } else {
        data.lastUser.loginId = loginId;
    }
    localStorage.setItem(ownidSdkStorageKey, JSON.stringify(data));
}

/**
 * getLastUser
 * @returns {string} - .
 */
function getLastUser() {
    const dataStr = localStorage.getItem(ownidSdkStorageKey);
    const data = dataStr ? JSON.parse(dataStr) : { lastUser: {} };
    return data.lastUser.loginId;
}

/**
 * onLogin
 * @param {Object} param0 - object
 * @param {boolean} [redirectToAccountPage=false] - .
 * @param {HTMLElement[]} [inputElements=[]] - .
 * @param {string} [widgetLocation] - .
 */
function onLogin({ token }, redirectToAccountPage = false, inputElements = [], widgetLocation = 'cartPage') {
    log('onLogin', widgetLocation);
    if (!token) {
        logError('Token is missing');
        return;
    }
    localStorage.removeItem(pageVisitLocalStorageKey);
    localStorage.removeItem(authNotificationShownStorageKey);
    $.ajax({
        url: window.ownIdUrls.setCookie,
        method: 'GET',
        data: {
            dwSid: token
        },
        success: function (data) {
            if (!data.success) {
                logError('Error on setCookie', data);
                return;
            }

            clearElementsErrors(inputElements);

            postAnalyticsEvent('signInSuccess', { signInType: 'ownID', source: widgetLocation });

            if (!data.basket) {
                if (redirectToAccountPage) {
                    window.location.assign(data.redirectUrl);
                } else {
                    localStorage.setItem(loggedInStorageKey, 'true');
                    window.location.reload();
                }
                return;
            }
            $.ajax({
                url: data.redirectUrl,
                method: 'POST',
                data: data,
                success: function (result) {
                    if (result.success) {
                        if (redirectToAccountPage) {
                            window.location.assign(result.redirectUrl);
                        } else {
                            localStorage.setItem(loggedInStorageKey, 'true');
                            window.location.reload();
                        }
                    }
                }
            });
        }
    });
}

/**
 * onRegister
 * @param {Object} data - object
 * @param {HTMLElement} loginIdElement - object
 * @param {HTMLElement} passwordElement - .
 * @param {string} widgetLocation - .
 */
function onRegister(data, loginIdElement, passwordElement, widgetLocation) {
    log('onRegister', widgetLocation);
    localStorage.removeItem(pageVisitLocalStorageKey);
    localStorage.removeItem(authNotificationShownStorageKey);
    localStorage.setItem(registrationDataLocalStorageKey, data.data);
    setLastUser(undefined);
    clearElementsErrors([loginIdElement, passwordElement]);
    postAnalyticsEvent('ownIDRegister', { source: widgetLocation });
    // eslint-disable-next-line no-param-reassign
    [...document.querySelectorAll('.toggle-password')]
        .filter(el => {
            const rect = el.getBoundingClientRect();
            return rect.x !== 0 && rect.y !== 0;
        })
        .forEach(el => {
            // eslint-disable-next-line no-param-reassign
            el.style.display = 'none';
        });
}

/**
 * setElementError
 * @param {HTMLElement} element - object
 * @param {string} message - .
 */
function setElementError(element, message) {
    element.classList.add('is-invalid');
    const errorElementId = `${element.id}-error`;
    var errorElement = document.getElementById(errorElementId);
    if (!errorElement) {
        errorElement = document.getElementById(errorElementId.replace(/(\*)-form-(\*)/, 'form-$2'));
        if (!errorElement) {
            errorElement = document.getElementById(
                `${element.id.split('-')
                    .slice(0, -1)
                    .join('-')}-error`
            );
            if (!errorElement) {
                errorElement = document.createElement('div');
                errorElement.id = errorElementId;
                errorElement.classList.add('invalid-feedback');
                const parentElement = element.parentElement;
                if (parentElement.lastChild === element) {
                    parentElement.appendChild(errorElement);
                } else {
                    parentElement.insertBefore(errorElement, element.nextSibling);
                }
            }
        }
    }
    errorElement.textContent = message;
    element.addEventListener('input', function () {
        clearElementsErrors([element]);
    });
}

/**
 * onLoginIdError
 * @param {string} type - .
 * @param {string} message - .
 * @param {Function} success - .
 * @param {Function|undefined} error - .
 */
function getMessage(type, message, success, error = undefined) {
    $.ajax({
        url: window.ownIdUrls.getMessage,
        method: 'GET',
        data: {
            type,
            message,
            siteAlias: isMKors() ? mkorsSiteAlias : defaultSiteAlias
        },
        success,
        error: error || (err => logError('Error on getMessage', err))
    });
}

/**
 * onLoginIdError
 * @param {HTMLElement} loginIdElement - object
 * @param {string} message - .
 * @param {string} type - .
 */
function onLoginIdError(loginIdElement, message, type) {
    log(`Login id error '${type}': ${message}`);
    getMessage(type, message, data => {
        setElementError(loginIdElement, data.message || message);
    }, error => {
        logError('Error on getMessage', error);
        setElementError(loginIdElement, message);
    });
}

/**
 * matchErrors
 * @param {HTMLElement} targetElement - object
 */
function matchElementErrorWithWidget(targetElement) {
    const observer = new MutationObserver(
        (mutationsList) => {
            for (let mutation of mutationsList) {
                if (mutation.type === 'attributes' && mutation.attributeName === 'class') {
                    const widgets = [...document.querySelectorAll(ownIdWidgetElementType)]
                        .filter(el => {
                            const rect = el.getBoundingClientRect();
                            return rect.x !== 0 && rect.y !== 0;
                        });
                    if (widgets.length > 0 && widgets[0].style) {
                        const widget = widgets[0];
                        if (targetElement.classList.contains('is-invalid')) {
                            if (originalBorderColor === null) {
                                originalBorderColor = widget.style.getPropertyValue(borderColorCssProperty) || '';
                            }
                            const computedStyle = window.getComputedStyle(targetElement);
                            widget.style.setProperty(borderColorCssProperty, computedStyle.borderColor);
                        } else if (originalBorderColor) {
                            widget.style.setProperty(borderColorCssProperty, originalBorderColor);
                        }
                    }
                }
            }
        }
    );
    observer.observe(targetElement, { attributes: true });
}

/**
 * setupWidget
 * @param {Object} object - object
 */
function setupWidget(
    {
        type,
        loginIdElementId,
        passwordElementId,
        targetElementId,
        onError,
        onLoginOverride,
        addOrStatement,
        widgetPosition,
        confirmPasswordContainer,
        additionalRequiredElementId,
        redirectToAccountPage,
        infoTooltip,
        widgetLocation
    } = {}
) {
    if (additionalRequiredElementId && !document.getElementById(additionalRequiredElementId)) {
        return;
    }

    const loginIdElement = document.getElementById(loginIdElementId);
    const targetElement = targetElementId ? document.getElementById(targetElementId) : undefined;
    const passwordElement = passwordElementId ? document.getElementById(passwordElementId) : undefined;
    if (loginIdElement && (passwordElement || targetElement)) {
        (passwordElement || targetElement).parentElement.style.zIndex = '2';
        window.ownid(type, {
            loginIdField: loginIdElement,
            targetField: targetElement,
            passwordField: passwordElement,
            confirmPasswordContainer,
            addOrStatement,
            widgetPosition,
            infoTooltip,
            resetFlowOnLoginIdChanged: true,
            onLogin: onLoginOverride || (props => onLogin(props, redirectToAccountPage, [loginIdElement, targetElement, passwordElement], widgetLocation)),
            onBeforeAuth: (user) => {
                if (widgetLocation) {
                    postAnalyticsEvent('ownIDClick', { source: widgetLocation });
                }
                return Promise.resolve(user);
            },
            onError: onError || (error => logError('Error', error)),
            onRegister: (data) => onRegister(data, loginIdElement, passwordElement || targetElement, widgetLocation),
            onLoginIdError: ({ message, type: errorType }) => {
                onLoginIdError(loginIdElement, message, errorType);
                return Promise.resolve({ hideTooltip: true });
            }
        });
        matchElementErrorWithWidget(passwordElement || targetElement);
    }
}

/**
 * tryGetParentElement
 * @param {HTMLElement} element - object
 * @returns {HTMLElement|undefined} - object
 */
function tryGetParentElement(element) {
    if (!element) {
        return undefined;
    }
    return element.parentElement;
}

/**
 * createLoginIdHolderIfNeeded
 * @param {string|undefined} loginId - object
 * @returns {undefined|string} - object
 */
function createLoginIdHolderIfNeeded(loginId) {
    if (!loginId || loginId.length === 0) {
        return undefined;
    }
    const loginIdHolder = document.createElement('input');
    loginIdHolder.type = 'hidden';
    loginIdHolder.value = loginId;
    loginIdHolder.id = 'ownid-loginIdHolder';
    document.body.appendChild(loginIdHolder);
    return loginIdHolder.id;
}

/**
 * activateReturningUsersPrompt
 */
function activateReturningUsersPrompt() {
    const action = 'activateReturningUsersPrompt';
    log('Activating returning users prompt');
    postAnalyticsEvent('ownIDPrompt', { promptType: 'returningUser' });
    const DELAY_MS = 1000;
    setTimeout(() => {
        window.ownid(action, {
            onLogin: resp => onLogin(resp, false, [], action),
            onBeforeAuth: (user) => {
                postAnalyticsEvent('ownIDClick', { source: action });
                return Promise.resolve(user);
            },
            onError: (errorMessage) => {
                logError(action, errorMessage);
            }
        })
            .then(res => log(action, res))
            .catch(err => logError(action, err));
    }, DELAY_MS);
}

/**
 * activateAccountRecoveryPrompt
 */
function activateAccountRecoveryPrompt() {
    const action = 'activateAccountRecoveryPrompt';
    log('Activating account recovery prompt');
    window.ownid(action, {
        onLogin: resp => onLogin(resp, true, [], action),
        onBeforeAuth: (user) => {
            postAnalyticsEvent('ownIDClick', { source: action });
            return Promise.resolve(user);
        },
        onError: (errorMessage) => {
            logError(action, errorMessage);
        }
    })
        .then(res => {
            log(action, res);
            if (!res.error) {
                postAnalyticsEvent('ownIDPrompt', { promptType: 'accountRecovery' });
            }
        })
        .catch(err => logError(action, err));
}

/**
 * showProductPageVisitAuthNotification
 */
function showAuthNotification() {
    log('Showing auth notification');
    getMessage('authNotification', '', data => {
        if (!data.messages || data.messages.length !== 3) {
            logError('Unexpected values on getMessage', data);
            return;
        }
        window.ownid('showAuthNotification', {
            onlyMobile: true,
            onlyReturningUsers: true,
            notification: {
                position: { anchor: 'top', offset: 100 },
                content: (lastUser) => ({
                    logoUrl: data.messages[0],
                    header: data.messages[1],
                    description: data.messages[2].replace('{loginId}', lastUser.loginId)
                })
            },
            onLogin: resp => onLogin(resp, false, [], 'productPageAuthNotification')
        })
            .then(({ error }) => {
                if (!error) {
                    postAnalyticsEvent('ownIDPrompt', { promptType: 'authNotification' });
                }
            });
    });
}

/**
 * setupReauth
 * @param {string} formSelector - .
 * @param {string|null} loginIdElementId - .
 * @param {string} passwordElementId - .
 * @param {string} widgetLocation - .
 */
function setupReauth(formSelector, loginIdElementId, passwordElementId, widgetLocation) {
    const form = document.querySelector(formSelector);
    if (!form) {
        return;
    }
    let loginIdElementIdActual = loginIdElementId;
    if (!loginIdElementIdActual) {
        if ((window.ownIdLoginId || '').length === 0) {
            return;
        }
        loginIdElementIdActual = createLoginIdHolderIfNeeded(window.ownIdLoginId);
        if (!loginIdElementIdActual) {
            return;
        }
    }
    setupWidget(
        {
            type: 'login',
            loginIdElementId: loginIdElementIdActual,
            passwordElementId,
            widgetLocation,
            infoTooltip: false,
            onLoginOverride: res => {
                const widgets = document.querySelectorAll('ownid-fingerprint-button-widget');
                widgets.forEach(widget => {
                    const w = widget;
                    w.shadowRoot.querySelector('ownid-spinner').style.display = 'none';
                    w.shadowRoot.querySelector('svg.check').style.display = 'unset';
                });
                document.getElementById(passwordElementId).value = res.metadata.authToken;
                form.action = form.action.replace('Account-', 'Ownid-');
            }
        }
    );
}

/**
 * setupWidgets
 */
function setupWidgets() {
    window.ownidOnLogin = onLogin;
    window.ownidSetup = setupWidget;

    if (window.ownIdLocale) {
        log('Setting locale', window.ownIdLocale);
        window.ownid('init', { language: window.ownIdLocale });
    }

    // guest checkout page
    setupWidget(
        {
            type: 'login',
            loginIdElementId: 'email',
            passwordElementId: 'password',
            additionalRequiredElementId: 'guest-customer',
            widgetLocation: 'checkoutPage'
        }
    );

    // login widget
    setupWidget(
        {
            type: 'login',
            loginIdElementId: 'login-form-email',
            passwordElementId: 'login-form-password',
            redirectToAccountPage: true,
            widgetLocation: 'topNotification'
        }
    );

    // login widget [MK custom]
    setupWidget(
        {
            type: 'login',
            loginIdElementId: 'page-login-form-email',
            passwordElementId: 'page-login-form-password',
            widgetLocation: 'loginPage'
        }
    );

    // register widget
    setupWidget(
        {
            type: 'register',
            loginIdElementId: 'registration-form-email',
            passwordElementId: 'registration-form-password',
            confirmPasswordContainer: [
                document.getElementById('registration-form-password-confirm'),
                document.querySelector('label[for=registration-form-password-confirm]'),
                document.getElementById('passwordValidationList'),
                document.querySelector('.dwfrm_profile_login_passwordconfirm'),
                document.getElementById('form-passwordError')
            ],
            redirectToAccountPage: true,
            widgetLocation: 'registerPage'
        }
    );

    // register post-buy
    const postBuyLoginIdElement = document.querySelector('#dwfrm_newPasswords .uneditable-emailID');
    setupWidget({
        type: 'register',
        loginIdElementId: createLoginIdHolderIfNeeded(postBuyLoginIdElement ? postBuyLoginIdElement.textContent : undefined),
        passwordElementId: 'newPassword',
        confirmPasswordContainer: [
            tryGetParentElement(document.querySelector('label[for=newPasswordConfirm]'))
        ],
        additionalRequiredElementId: 'dwfrm_newPasswords',
        widgetLocation: 'postBuy'
    });

    if (window.ownidShowReturningUserPrompt) {
        activateReturningUsersPrompt();
    }

    if (window.ownidShowReturningUserPromptOnSecondVisit && getLastUser()) {
        if (!localStorage.getItem(pageVisitLocalStorageKey)) {
            log('Setting page visited');
            localStorage.setItem(pageVisitLocalStorageKey, 'true');
        } else {
            activateReturningUsersPrompt();
            localStorage.removeItem(pageVisitLocalStorageKey);
        }
    }

    if (window.ownidShowAuthNotification && getLastUser()) {
        if (!localStorage.getItem(authNotificationShownStorageKey)) {
            showAuthNotification();
            localStorage.setItem(authNotificationShownStorageKey, 'true');
        }
    }

    if (window.ownidShowAccountRecovery) {
        setupWidget(
            {
                type: 'recover',
                loginIdElementId: 'reset-password-email',
                targetElementId: 'reset-password-email',
                addOrStatement: false,
                widgetPosition: 'end',
                redirectToAccountPage: true,
                infoTooltip: false,
                widgetLocation: 'passwordReset'
            }
        );
        activateAccountRecoveryPrompt();
    }

    if (window.ownidEnableProfileUpdateWidgets) {
        setupReauth(
            'form[name=dwfrm_profile]',
            'email',
            'password',
            'updateProfile'
        );

        setupReauth(
            'form[name=dwfrm_profile]',
            'edit-form-email',
            'edit-form-password',
            'updateProfile'
        );

        setupReauth(
            'form[name=dwfrm_profile]',
            null,
            'currentPassword',
            'updatePassword'
        );
    }
}

/**
 * checkRegistrationData
 */
function checkRegistrationData() {
    const registrationData = localStorage.getItem(registrationDataLocalStorageKey);
    if (!registrationData) {
        return;
    }
    log('Setting registration data');
    localStorage.removeItem(registrationDataLocalStorageKey);
    $.ajax({
        url: window.ownIdUrls.setData,
        method: 'POST',
        data: {
            data: registrationData
        },
        success: function (data) {
            if (data.success && data.email) {
                setLastUser(data.email);
            }
            if (!data.success) {
                setLastUser(undefined);
            }
        },
        error: function (error) {
            logError('Error on setData', error);
            setLastUser(undefined);
        }
    });
}

/**
 * checkForLoginHintInQueryParams
 */
function checkForLoginHintInQueryParams() {
    const loginHintQueryString = 'loginHint';
    const url = new URL(window.location.href);
    if (!url.searchParams.has(loginHintQueryString) || url.searchParams.get(loginHintQueryString) !== 'true') {
        return;
    }
    log('Detected login hint', url);
    localStorage.setItem(enrollCredentialLocalStorageKey, 'true');
    url.searchParams.delete(loginHintQueryString);
    window.history.replaceState(null, '', url.toString());
}

/**
 * setupEnrollCredential
 */
function setupEnrollCredential() {
    checkForLoginHintInQueryParams();

    if (localStorage.getItem(enrollCredentialLocalStorageKey)) {
        log('Triggering enroll credential');
        $.ajax({
            url: window.ownIdUrls.getJwt,
            method: 'GET',
            success: function (data) {
                if (data.jwt) {
                    window.ownid.enrollCredential(data.jwt)
                        .then(res => log('Enroll credentials completed', res))
                        .catch(err => logError('Enroll credentials failed', err));
                }
            },
            complete: function () {
                localStorage.removeItem(enrollCredentialLocalStorageKey);
            }
        });
    }
}

/**
 * toast
 * @param {string} message - .
 */
function toast(message) {
    window.ownid('toast', {
        dismissTimeout: 3000,
        status: 'success',
        position: {
            offset: 100
        },
        content: {
            header: !message || message.length === 0 ? undefined : message
        }
    });
}

/**
 * showLoggedInPromptIfNeeded
 */
function showLoggedInPromptIfNeeded() {
    if (!localStorage.getItem(loggedInStorageKey)) {
        return;
    }
    getMessage('signedIn', '', data => toast(data.message));
    localStorage.removeItem(loggedInStorageKey);
}

/**
 * waitForIt
 * @param {Function} it - .
 * @returns {Promise<void>} - .
 */
function waitForIt(it) {
    return new Promise((resolve) => {
        const interval = setInterval(function () {
            if (it()) {
                clearInterval(interval);
                resolve();
            }
        }, 10);
    });
}

/**
 * initOwnIDWidgets
 */
function initOwnIDWidgets() {
    log(`cartridge version ${VERSION}`);
    const logoElementSelector = 'div.navbar-header.brand';

    waitForIt(() => document.querySelector(logoElementSelector))
        .then(() => {
            if (window.ownIdEnabled === false) {
                return;
            }
            waitForIt(() => window.ownid && window.ownid.sdk)
                .then(() => {
                    if (window.ownidLoaded) {
                        return;
                    }
                    window.ownidLoaded = true;
                    setupWidgets();
                    checkRegistrationData();
                    setupEnrollCredential();
                    showLoggedInPromptIfNeeded();
                });
        });
}

if (window.location.hash.toLowerCase() === '#ownid_clear_rup') {
    log('Clearing RUP');
    const storeObj = JSON.parse(window.localStorage.getItem(ownidSdkStorageKey) || '{}');
    delete storeObj['ownid-ac-block'];
    window.localStorage.setItem(ownidSdkStorageKey, JSON.stringify(storeObj));
}

module.exports = {
    initOwnIDWidgets: initOwnIDWidgets()
};
