'use strict';

var server = require('server');
const Encoding = require('dw/crypto/Encoding');
const Bytes = require('dw/util/Bytes');
var csrfProtection = require('*/cartridge/scripts/middleware/csrf');

/**
 * setHttpCookie
 * @param {Request} req - The HTTP request object
 * @returns {boolean} - The success of setting the cookie
 */
function setHttpCookie(req) {
    var Cookie = require('dw/web/Cookie');
    var Logger = require('dw/system/Logger');

    // dwsid is the session token, it must be set to complete the login flow
    var isDwSidTokenInRequest = Object.hasOwnProperty.call(req.querystring, 'dwSid');
    if (!isDwSidTokenInRequest) {
        return false;
    }
    var dwSidToken = req.querystring.dwSid;
    try {
        var newCookie = new Cookie('dwsid', dwSidToken);
        newCookie.setPath('/');
        response.addHttpCookie(newCookie);
        return true;
    } catch (e) {
        Logger.error('Error while setting dwSid :' + e.message);
    }
    return false;
}

/**
 * Ownid-SetCookie : This endpoint is called to set dwSid in the Cookie and create basket data for merge basket
 * @name Base/Ownid-SetCookie
 * @function
 * @memberof Ownid
 * @param {returns} - json
 * @param {serverfunction} - get
 */
server.get('SetCookie', function (req, res, next) {
    var URLUtils = require('dw/web/URLUtils');
    var StringUtils = require('dw/util/StringUtils');
    var BasketMgr = require('dw/order/BasketMgr');
    var collections = require('*/cartridge/scripts/util/collections');

    var currentBasket = BasketMgr.getCurrentBasket();
    var allProductLineItems = currentBasket ? currentBasket.allProductLineItems : [];
    // set dwsid in cookie
    var cookieResponse = setHttpCookie(req);
    var allProductData = {};
    allProductData.data = [];
    allProductData.coupons = [];
    // fetching all the product from the Basket
    if (allProductLineItems.length > 0) {
        collections.forEach(allProductLineItems, function (currentProduct) {
            var parentIndex;
            if (currentProduct.optionID) {
                parentIndex = allProductData.data.findIndex(product => product.productId === currentProduct.parent.productID);
                var optionProductJson = {
                    optionId: currentProduct.optionID,
                    selectedValueId: currentProduct.productID
                };
                allProductData.data[parentIndex].options.push(optionProductJson);
            } else if (currentProduct.bundledProductLineItem) {
                parentIndex = allProductData.data.findIndex(product => product.productId === currentProduct.parent.productID);
                var childProductJson = {
                    pid: currentProduct.productID,
                    quantity: currentProduct.quantityValue
                };
                allProductData.data[parentIndex].childProducts.push(childProductJson);
            } else {
                var currentProductData = {};
                currentProductData.productId = currentProduct.productID;
                currentProductData.productQuantity = currentProduct.quantityValue;
                currentProductData.options = [];
                currentProductData.childProducts = [];
                allProductData.data.push(currentProductData);
            }
        });
        var couponLineItems = currentBasket.couponLineItems;
        if (couponLineItems.length > 0) {
            collections.forEach(couponLineItems, function (currentCoupon) {
                allProductData.coupons.push(currentCoupon.couponCode);
            });
        }
    } else {
        res.json({
            success: cookieResponse,
            redirectUrl: URLUtils.abs('Account-Show')
                .toString(),
            basket: false
        });
        return next();
    }
    // Encoding basket data to send on OwnId-MergeBasket route for merging
    var allProductDataBase64 = '';
    try {
        var allProductDataString = JSON.stringify(allProductData);
        allProductDataBase64 = StringUtils.encodeBase64(allProductDataString);
    } catch (error) {
        res.json({
            success: false
        });
        return next();
    }
    res.json({
        success: cookieResponse,
        redirectUrl: URLUtils.abs('Ownid-MergeBasket')
            .toString(),
        basket: allProductDataBase64
    });
    return next();
});


/**
 * Ownid-MergeBasket : This endpoint is called to merge the guest basket with the logged-in customer's basket
 * @name Base/Ownid-MergeBasket
 * @function
 * @memberof Ownid
 * @param {returns} - json
 * @param {serverfunction} - post
 */
server.post('MergeBasket', function (req, res, next) {
    var URLUtils = require('dw/web/URLUtils');
    var StringUtils = require('dw/util/StringUtils');
    var BasketMgr = require('dw/order/BasketMgr');
    var Transaction = require('dw/system/Transaction');
    var collections = require('*/cartridge/scripts/util/collections');
    var cartHelper = require('*/cartridge/scripts/cart/cartHelpers');
    var basketCalculationHelpers = require('*/cartridge/scripts/helpers/basketCalculationHelpers');

    var currentBasket = BasketMgr.getCurrentOrNewBasket();
    try {
        var guestBasket = req.form.basket;
        // Decoding basket data
        var guestBasketDataString = StringUtils.decodeBase64(guestBasket);
        var guestBasketData = JSON.parse(guestBasketDataString);
        Transaction.wrap(function () {
            if (currentBasket.allProductLineItems.length === 0) {
                collections.forEach(currentBasket.couponLineItems, function (currentCoupon) {
                    currentBasket.removeCouponLineItem(currentCoupon);
                });
            }
            if (guestBasketData.data.length > 0 && currentBasket.allProductLineItems.length > 0) {
                var allProductLineItems = currentBasket.allProductLineItems;
                collections.forEach(allProductLineItems, function (currentProduct) {
                    currentBasket.removeProductLineItem(currentProduct);
                });
                collections.forEach(currentBasket.couponLineItems, function (currentCoupon) {
                    currentBasket.removeCouponLineItem(currentCoupon);
                });
            }
            for (var currentGuestProduct of guestBasketData.data) {
                cartHelper.addProductToCart(
                    currentBasket,
                    currentGuestProduct.productId,
                    currentGuestProduct.productQuantity,
                    currentGuestProduct.childProducts,
                    currentGuestProduct.options
                );
            }
            if (guestBasketData.data.length > 0) {
                for (var currentCoupon of guestBasketData.coupons) {
                    currentBasket.createCouponLineItem(currentCoupon, true);
                }
            }
            basketCalculationHelpers.calculateTotals(currentBasket);
        });
    } catch (error) {
        res.json({
            success: false
        });
        return next();
    }
    res.json({
        success: true,
        redirectUrl: URLUtils.abs('Account-Show')
            .toString()
    });
    return next();
});

server.post('Data', function (req, res, next) {
    var CustomerMgr = require('dw/customer/CustomerMgr');
    var Transaction = require('dw/system/Transaction');

    if (!req.currentCustomer.profile) {
        res.json({ success: false, message: 'User not authenticated' });
        return next();
    }

    var currentCustomer = CustomerMgr.getCustomerByCustomerNumber(req.currentCustomer.profile.customerNo);

    var ownIdData = req.form.data;
    if (!ownIdData) {
        res.json({ success: false, message: 'No data provided' });
        return next();
    }

    try {
        Transaction.wrap(function () {
            var customerProfile = currentCustomer.getProfile();
            customerProfile.custom.ownIdData = ownIdData;
        });

        res.json({ success: true, message: 'OwnID data set successfully', email: req.currentCustomer.profile.email });
    } catch (e) {
        res.setStatusCode(500);
        res.json({ success: false, message: 'Error updating OwnID data', error: e.message });
    }

    return next();
});

/**
 * encodeBase64Url
 * @param {Object} obj - The object to be encoded.
 * @returns {string} The Base64 URL-encoded string.
 */
function encodeBase64Url(obj) {
    return Encoding.toBase64URL(new Bytes(JSON.stringify(obj)));
}

server.get('Jwt', function (req, res, next) {
    const Signature = require('dw/crypto/Signature');
    const Site = require('dw/system/Site');

    if (!req.currentCustomer.profile) {
        res.json({
            error: 'Unauthorized'
        });
        return next();
    }

    const currentSite = Site.getCurrent();
    var signingKey = currentSite.getCustomPreferenceValue('signingKey');
    if (!signingKey) {
        res.json({
            error: 'No signing key'
        });
        return next();
    }
    signingKey = signingKey
        .replace(/-----BEGIN RSA PRIVATE KEY-----/g, '')
        .replace(/-----END RSA PRIVATE KEY-----/g, '')
        .trim();

    const env = currentSite.getCustomPreferenceValue('env')
        .getValue()
        .toLowerCase();
    const serverEnv = env === 'prod' ? '' : '.' + env;
    const appUrl = currentSite.getCustomPreferenceValue('appId') + '.server' + serverEnv + '.ownid.com';

    const base64UrlHeader = encodeBase64Url({
        alg: 'RS256',
        typ: 'JWT',
        kid: appUrl
    });
    const tokenTtlSec = 60 * 60;
    const base64UrlPayload = encodeBase64Url({
        sub: req.currentCustomer.profile.email,
        iss: 'OwnID',
        aud: 'OwnID',
        iat: Math.floor(Date.now() / 1000),
        exp: Math.floor(Date.now() / 1000) + tokenTtlSec,
        jti: Math.random()
            .toString(36)
            .substring(7)
    });

    const signatureInput = base64UrlHeader + '.' + base64UrlPayload;

    const signature = new Signature();
    const signatureBytes = signature.signBytes(new Bytes(signatureInput), signingKey, 'SHA256withRSA');
    const base64UrlSignature = Encoding.toBase64URL(signatureBytes);

    res.json({
        jwt: signatureInput + '.' + base64UrlSignature
    });
    return next();
});

server.get('Message', function (req, res, next) {
    var Resource = require('dw/web/Resource');

    var messageType = req.querystring.type;
    var defaultMessage = req.querystring.message;
    var siteAlias = req.querystring.siteAlias || 'default';
    var message = '';
    var messages = [];
    switch (messageType) {
        case 'signedIn':
            message = Resource.msg('msg.login.' + siteAlias, 'ownid', defaultMessage);
            break;
        case 'authNotification':
            messages = [
                Resource.msg('msg.authNotification.logo.' + siteAlias, 'ownid', null),
                Resource.msg('msg.authNotification.header.' + siteAlias, 'ownid', null),
                Resource.msg('msg.authNotification.description.' + siteAlias, 'ownid', null)
            ];
            break;
        // see `LoginIdErrorType` at web-sdk: `i-widget.interfaces.ts`
        case 'empty':
            message = Resource.msg('error.message.missing.email.required', 'forms.properties', defaultMessage);
            break;
        case 'invalidFormat':
            message = Resource.msg('emailsignup.error.pattern', 'forms.properties', defaultMessage);
            break;
        case 'accountNotFound':
        case 'accountIsBlocked':
        default:
            message = defaultMessage;
            break;
    }
    res.json({
        message: message,
        messages: messages
    });
    return next();
});

/**
 * validateJwt
 * @param {string} token - .
 * @returns {{success: boolean}|{error: string}} - .
 */
function validateJwt(token) {
    const Site = require('dw/system/Site');
    const HTTPClient = require('dw/net/HTTPClient');
    const Signature = require('dw/crypto/Signature');

    const currentSite = Site.getCurrent();
    const env = currentSite.getCustomPreferenceValue('env')
        .getValue()
        .toLowerCase();
    const envPart = env === 'prod' ? '' : '.' + env;
    const pubKeyUrl = 'https://' + currentSite.getCustomPreferenceValue('appId') + '.server' + envPart + '.ownid.com/oidc/public/default';

    const httpClient = new HTTPClient();
    httpClient.open('GET', pubKeyUrl);
    httpClient.send();
    if (httpClient.statusCode !== 200) {
        return { error: 'Failed to fetch public key for app from ' + pubKeyUrl };
    }
    const pubKey = httpClient.text;

    const tokenParts = token.split('.');
    if (tokenParts.length !== 3) {
        return { error: 'Invalid JWT format.' };
    }

    const contentToVerify = tokenParts[0] + '.' + tokenParts[1];
    const contentToVerifyInBytes = new Bytes(contentToVerify);
    const jwtSignature = tokenParts[2];
    const jwtSignatureInBytes = Encoding.fromBase64(jwtSignature);

    const sig = new Signature();
    const verified = sig.verifyBytesSignature(jwtSignatureInBytes, contentToVerifyInBytes, pubKey, 'SHA256withRSA');
    if (verified) {
        return { success: true };
    }
    return { error: 'JWT signature verification failed.' };
}

/**
 * Adapted from Account-SaveProfile : The Account-SaveProfile endpoint is the endpoint that gets hit when a shopper has edited their profile
 * @name Base/Account-SaveProfile
 * @function
 * @memberof Account
 * @param {middleware} - server.middleware.https
 * @param {middleware} - csrfProtection.validateAjaxRequest
 * @param {httpparameter} - dwfrm_profile_customer_firstname - Input field for the shoppers's first name
 * @param {httpparameter} - dwfrm_profile_customer_lastname - Input field for the shopper's last name
 * @param {httpparameter} - dwfrm_profile_customer_phone - Input field for the shopper's phone number
 * @param {httpparameter} - dwfrm_profile_customer_email - Input field for the shopper's email address
 * @param {httpparameter} - dwfrm_profile_customer_emailconfirm - Input field for the shopper's email address
 * @param {httpparameter} - dwfrm_profile_login_password  - Input field for the shopper's password
 * @param {httpparameter} - csrf_token - hidden input field CSRF token
 * @param {category} - sensititve
 * @param {returns} - json
 * @param {serverfunction} - post
 */
server.post(
    'SaveProfile',
    server.middleware.https,
    csrfProtection.validateAjaxRequest,
    function (req, res, next) {
        var Transaction = require('dw/system/Transaction');
        var CustomerMgr = require('dw/customer/CustomerMgr');
        var Resource = require('dw/web/Resource');
        var URLUtils = require('dw/web/URLUtils');
        var accountHelpers = require('*/cartridge/scripts/helpers/accountHelpers');

        var formErrors = require('*/cartridge/scripts/formErrors');

        var profileForm = server.forms.getForm('profile');

        profileForm.login.password.valid = true;
        if (profileForm.customer.email.value.toLowerCase()
            !== profileForm.customer.emailconfirm.value.toLowerCase()) {
            profileForm.valid = false;
            profileForm.customer.email.valid = false;
            profileForm.customer.emailconfirm.valid = false;
            profileForm.customer.emailconfirm.error =
                Resource.msg('error.message.mismatch.email', 'forms', null);
            res.json({
                success: false,
                fields: formErrors.getFormErrors(profileForm)
            });
            return next();
        }

        const verificationResult = validateJwt(profileForm.login.password.value);
        if (!verificationResult.success) {
            profileForm.login.password.valid = false;
            profileForm.login.password.error = verificationResult.error;
            res.json({
                success: false,
                fields: formErrors.getFormErrors(profileForm)
            });
            return next();
        }

        var result = {
            firstName: profileForm.customer.firstname.value,
            lastName: profileForm.customer.lastname.value,
            phone: profileForm.customer.phone.value,
            email: profileForm.customer.email.value,
            confirmEmail: profileForm.customer.emailconfirm.value,
            profileForm: profileForm
        };
        res.setViewData(result);
        this.on('route:BeforeComplete', function (req, res) { // eslint-disable-line no-shadow
            var formInfo = res.getViewData();
            var customer = CustomerMgr.getCustomerByCustomerNumber(
                req.currentCustomer.profile.customerNo
            );
            var profile = customer.getProfile();

            delete formInfo.password;
            delete formInfo.confirmEmail;

            Transaction.wrap(function () {
                profile.setFirstName(formInfo.firstName);
                profile.setLastName(formInfo.lastName);
                profile.setEmail(formInfo.email);
                profile.setPhoneHome(formInfo.phone);
            });

            accountHelpers.sendAccountEditedEmail(customer.profile);

            delete formInfo.profileForm;
            delete formInfo.email;

            res.json({
                success: true,
                redirectUrl: URLUtils.url('Account-Show')
                    .toString()
            });
        });
        return next();
    }
);


/**
 * Adapted from Account-SavePassword : The Account-SavePassword endpoint is the endpoit that handles changing the shopper's password
 * @name Base/Account-SavePassword
 * @function
 * @memberof Account
 * @param {middleware} - server.middleware.https
 * @param {middleware} - csrfProtection.validateAjaxRequest
 * @param {httpparameter} - dwfrm_profile_login_currentpassword - Input field for the shopper's current password
 * @param {httpparameter} - dwfrm_profile_login_newpasswords_newpassword - Input field for the shopper's new password
 * @param {httpparameter} - dwfrm_profile_login_newpasswords_newpasswordconfirm - Input field for the shopper to confirm their new password
 * @param {httpparameter} - csrf_token - hidden input field CSRF token
 * @param {category} - sensitive
 * @param {returns} - json
 * @param {serverfunction} - post
 */
server.post(
    'SavePassword',
    server.middleware.https,
    csrfProtection.validateAjaxRequest,
    function (req, res, next) {
        var Transaction = require('dw/system/Transaction');
        var CustomerMgr = require('dw/customer/CustomerMgr');
        var Resource = require('dw/web/Resource');
        var URLUtils = require('dw/web/URLUtils');

        var formErrors = require('*/cartridge/scripts/formErrors');

        var profileForm = server.forms.getForm('profile');
        var newPasswords = profileForm.login.newpasswords;

        profileForm.login.currentpassword.valid = true;
        if (newPasswords.newpassword.value !== newPasswords.newpasswordconfirm.value) {
            profileForm.valid = false;
            newPasswords.newpassword.valid = false;
            newPasswords.newpasswordconfirm.valid = false;
            newPasswords.newpasswordconfirm.error =
                Resource.msg('error.message.mismatch.newpassword', 'forms', null);
            res.json({
                success: false,
                fields: formErrors.getFormErrors(profileForm)
            });
            return next();
        }

        const verificationResult = validateJwt(profileForm.login.currentpassword.value);
        if (!verificationResult.success) {
            profileForm.login.currentpassword.valid = false;
            profileForm.login.currentpassword.error = verificationResult.error;
            res.json({
                success: false,
                fields: formErrors.getFormErrors(profileForm)
            });
            return next();
        }

        var result = {
            newPassword: newPasswords.newpassword.value,
            newPasswordConfirm: newPasswords.newpasswordconfirm.value,
            profileForm: profileForm
        };
        res.setViewData(result);
        this.on('route:BeforeComplete', function () { // eslint-disable-line no-shadow
            var formInfo = res.getViewData();
            var customer = CustomerMgr.getCustomerByCustomerNumber(
                req.currentCustomer.profile.customerNo
            );
            var status;
            Transaction.wrap(function () {
                status = customer.profile.credentials.setPassword(
                    formInfo.newPassword,
                    '',
                    false
                );
            });
            if (status.error) {
                if (!CustomerMgr.isAcceptablePassword(newPasswords.newpassword.value)) {
                    formInfo.profileForm.login.newpasswords.newpassword.valid = false;
                    formInfo.profileForm.login.newpasswords.newpassword.error =
                        Resource.msg('error.message.password.constraints.not.matched', 'forms', null);
                }

                delete formInfo.currentPassword;
                delete formInfo.newPassword;
                delete formInfo.newPasswordConfirm;
                delete formInfo.profileForm;

                res.json({
                    success: false,
                    fields: formErrors.getFormErrors(profileForm)
                });
            } else {
                delete formInfo.currentPassword;
                delete formInfo.newPassword;
                delete formInfo.newPasswordConfirm;
                delete formInfo.profileForm;

                res.json({
                    success: true,
                    redirectUrl: URLUtils.url('Account-Show')
                        .toString()
                });
            }
        });
        return next();
    }
);

module.exports = server.exports();
